import pytest
import pymongo

import itertools
import datetime
import uuid

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.statestore.stores.dto import CombinerDTO


@pytest.fixture
def test_combiners():
    start_date = datetime.datetime(2021, 1, 4, 1, 2, 4)
    combiner1 = CombinerDTO(combiner_id=str(uuid.uuid4()), name="test_combiner1",
                  parent="localhost", ip="123:13:12:2", fqdn="", port=8080,
                  updated_at=start_date - datetime.timedelta(days=52), address="test_address")
    combiner2 = CombinerDTO(combiner_id=str(uuid.uuid4()), name="test_combiner2",
                  parent="localhost", ip="123:13:12:2", fqdn="", port=8080,
                  updated_at=start_date - datetime.timedelta(days=12), address="test_address") 
    combiner3 = CombinerDTO(combiner_id=str(uuid.uuid4()), name="test_combiner3",
                    parent="localhost", ip="123:13:12:5", fqdn="", port=8080,
                    updated_at=start_date - datetime.timedelta(days=322), address="test_address")
    combiner4 = CombinerDTO(combiner_id=str(uuid.uuid4()), name="test_combiner4",
                    parent="localhost", ip="123:13:12:4", fqdn="", port=8080,
                    updated_at=start_date - datetime.timedelta(days=23), address="test_address")
    combiner5 = CombinerDTO(combiner_id=str(uuid.uuid4()), name="test_combiner5",
                    parent="localhost", ip="123:13:12:3", fqdn="", port=8080,
                    updated_at=start_date - datetime.timedelta(days=22), address="test_address")
    combiner6 = CombinerDTO(combiner_id=str(uuid.uuid4()), name="test_combiner6",
                    parent="localhost", ip="123:13:12:3", fqdn="", port=8080,
                    updated_at=start_date - datetime.timedelta(days=24), address="test_address")
    combiner7 = CombinerDTO(combiner_id=str(uuid.uuid4()), name="test_combiner8",
                    parent="localhost", ip="123:13:12:3", fqdn="", port=8080,
                    updated_at=start_date - datetime.timedelta(days=42), address="test_address")
    combiner8 = CombinerDTO(combiner_id=str(uuid.uuid4()), name="test_combiner7",
                    parent="localhost", ip="123:13:12:2", fqdn="", port=8080,
                    updated_at=start_date - datetime.timedelta(days=12), address="test_address1")
    return [combiner1, combiner2, combiner3, combiner4, combiner5, combiner6, combiner7, combiner8]

@pytest.fixture
def test_combiner():
    start_date = datetime.datetime(2021, 1, 4, 1, 2, 4)
    combiner = CombinerDTO(name="test_combiner",
                  parent="localhost", ip="123:13:12:2", fqdn="", port=8080,
                  updated_at=start_date - datetime.timedelta(days=52), address="test_address")
    return combiner

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_combiners):
    for c in test_combiners:
        mongo_connection.combiner_store.add(c)
        postgres_connection.combiner_store.add(c)
        sql_connection.combiner_store.add(c)

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    for c in test_combiners:
        mongo_connection.combiner_store.delete(c.combiner_id)
        postgres_connection.combiner_store.delete(c.combiner_id)
        sql_connection.combiner_store.delete(c.combiner_id)
    

@pytest.fixture
def options():
    sorting_keys = (None, 
                    "name",
                    "committed_at",
                    "updated_at",
                    "invalid_key") 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"ip":"123:13:12:3"}, {"address":"test_address1"}, {"fqdn":"", "name":"test_combiner77"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestCombinerStore:

    def test_add_update_delete(self, db_connection: DatabaseConnection, test_combiner:CombinerDTO):
        test_combiner.check_validity()

        # Add a combiner and check that we get the added combiner back
        name = test_combiner.name

        read_combiner1 = db_connection.combiner_store.add(test_combiner)
        assert isinstance(read_combiner1.combiner_id, str)
        read_combiner1_dict = read_combiner1.to_dict()
        combiner_id = read_combiner1_dict["combiner_id"]
        del read_combiner1_dict["combiner_id"]
        del read_combiner1_dict["committed_at"]

        test_combiner_dict = test_combiner.to_dict()
        del test_combiner_dict["combiner_id"]
        del test_combiner_dict["committed_at"]

        assert read_combiner1_dict == test_combiner_dict

        # Assert we get the same combiner back
        read_combiner2 = db_connection.combiner_store.get(combiner_id)
        assert read_combiner2 is not None
        assert read_combiner2.to_dict() == read_combiner1.to_dict()

        # Assert we get the same combiner back by name
        read_combiner3 = db_connection.combiner_store.get_by_name(name)
        assert read_combiner3 is not None
        assert read_combiner3.to_dict() == read_combiner1.to_dict()

        # Delete the combiner and check that it is deleted
        success = db_connection.combiner_store.delete(combiner_id)
        assert success == True

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                gathered_combiners = db_1.combiner_store.list(*opt, **kwargs)
                count =  db_1.combiner_store.count(**kwargs)
                
                gathered_combiners2 = db_2.combiner_store.list(*opt, **kwargs)
                count2 = db_2.combiner_store.count(**kwargs)

                assert(count == count2)
                assert len(gathered_combiners) == len(gathered_combiners2)

                for i in range(len(gathered_combiners)):
                    assert gathered_combiners2[i].combiner_id == gathered_combiners[i].combiner_id
                        

                        
    



from datetime import datetime
from typing import Optional

from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, Field, PrimaryID


class ClientDTO(BaseDTO):
    """Client data transfer object."""

    client_id: Optional[str] = PrimaryID(None)
    name: str = Field(None)
    combiner: str = Field(None)
    combiner_preferred: Optional[str] = Field(None)
    ip: Optional[str] = Field(None)
    status: str = Field(None)
    last_seen: datetime = Field(None)
    package: Optional[str] = Field(None)



import pytest
import pymongo

import itertools
import datetime
import uuid

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.statestore.stores.dto.package import PackageDTO

@pytest.fixture
def test_packages():
        package1 = PackageDTO(package_id=str(uuid.uuid4()), description="Test package1", file_name="test_package.tar.gz", helper="numpyhelper", name="test_package1")
        package2 = PackageDTO(package_id=str(uuid.uuid4()), description="Test package1", file_name="test_package.tar.gz", helper="numpyhelper", name="test_package2")
        package3 = PackageDTO(package_id=str(uuid.uuid4()), description="Test package1", file_name="test_package.tar.gz", helper="numpyhelper", name="test_package5")
        package4 = PackageDTO(package_id=str(uuid.uuid4()), description="Test package2", file_name="test_package.tar.gz", helper="numpyhelper", name="test_package4")
        package5 = PackageDTO(package_id=str(uuid.uuid4()), description="Test package2", file_name="test_package.tar.gz", helper="numpyhelper", name="test_package7")
        package6 = PackageDTO(package_id=str(uuid.uuid4()), description="Test package2", file_name="test_package.tar.gz", helper="numpyhelper", name="test_package3")
        return [package1, package2, package3, package4, package5, package6]

@pytest.fixture
def test_package():
    return PackageDTO(description="Test package1", file_name="test_package.tar.gz", helper="numpyhelper", name="test_package1")

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_packages):
    for c in test_packages:
        mongo_connection.package_store.add(c)
        postgres_connection.package_store.add(c)
        sql_connection.package_store.add(c)

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    for c in test_packages:
        mongo_connection.package_store.delete(c.package_id)
        postgres_connection.package_store.delete(c.package_id)
        sql_connection.package_store.delete(c.package_id)



@pytest.fixture
def options():
    sorting_keys = (None, 
                    "name",
                    "committed_at",
                    "invalid_key"
                    ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"description":"Test package2"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestPackageStore:
    
    def test_add_update_delete(self, db_connection: DatabaseConnection, test_package: PackageDTO):
        test_package.check_validity()

        # Add a package and check that we get the added package back
        read_package1 = db_connection.package_store.add(test_package)
        assert isinstance(read_package1.package_id, str)
        assert isinstance(read_package1.committed_at, datetime.datetime)
        assert isinstance(read_package1.storage_file_name, str)

        read_package1_dict = read_package1.to_dict()
        package_id = read_package1_dict["package_id"]
        del read_package1_dict["package_id"]
        del read_package1_dict["committed_at"]
        del read_package1_dict["storage_file_name"]

        test_package_dict = test_package.to_dict()
        del test_package_dict["package_id"]
        del test_package_dict["committed_at"]
        del test_package_dict["storage_file_name"]

        assert read_package1_dict == test_package_dict

        # Assert we get the same package back
        read_package2 = db_connection.package_store.get(package_id)
        assert read_package2 is not None
        assert read_package2.to_dict() == read_package1.to_dict()    

        # Delete the package and check that it is deleted
        success = db_connection.package_store.delete(package_id)
        assert success == True
    
    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                gathered_packages = db_1.package_store.list(*opt, **kwargs)
                count = db_1.package_store.count(**kwargs)
                
                gathered_packages2 = db_2.package_store.list(*opt, **kwargs)
                count2 = db_2.package_store.count(**kwargs)
                
                assert count == count2
                assert len(gathered_packages) == len(gathered_packages2)

                for i in range(len(gathered_packages)):
                    assert gathered_packages2[i].package_id == gathered_packages[i].package_id
                    
                    



import pytest
import pymongo

import datetime
import uuid
import itertools

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.statestore.stores.dto.prediction import PredictionDTO




@pytest.fixture
def test_predictions():
    pred1 = PredictionDTO(prediction_id=str(uuid.uuid4()), correlation_id="test_correlation_id1", data="test_data1", model_id=None,
                          receiver={"name":"test_receiver_name1", "role":"test_receiver_role1"}, sender={"name":"test_sender_name1",
                          "role":"test_sender_role1"}, timestamp="2024-13")
    pred2 = PredictionDTO(prediction_id=str(uuid.uuid4()), correlation_id="test_correlation_id2", data="test_data1", model_id=None,
                            receiver={"name":"test_receiver_name2", "role":"test_receiver_role1"}, sender={"name":"test_sender_name1",
                            "role":"test_sender_role1"}, timestamp="2024-11")
    pred3 = PredictionDTO(prediction_id=str(uuid.uuid4()), correlation_id="test_correlation_id3", data="test_data1", model_id=None,
                            receiver={"name":"test_receiver_name3", "role":"test_receiver_role1"}, sender={"name":"test_sender_name2",
                            "role":"test_sender_role1"}, timestamp="2024-12")
    pred4 = PredictionDTO(prediction_id=str(uuid.uuid4()), correlation_id="test_correlation_id4", data="test_data1", model_id=None,
                            receiver={"name":"test_receiver_name4", "role":"test_receiver_role1"}, sender={"name":"test_sender_name2",
                            "role":"test_sender_role1"}, timestamp="2024-16")
    return [pred1, pred2, pred3, pred4]



@pytest.fixture
def test_prediction():
    prediction = PredictionDTO()
    prediction.correlation_id = "test_correlation_id1"
    prediction.data = "test_data1"
    prediction.model_id = None
    prediction.receiver.name = "test_receiver_name1"
    prediction.receiver.role = "test_receiver_role1"
    prediction.sender.name = "test_sender_name1"
    prediction.sender.role = "test_sender_role1"
    prediction.timestamp = "2024-13"
    return prediction


@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_predictions):
    for c in test_predictions:
        mongo_connection.prediction_store.add(c)
        postgres_connection.prediction_store.add(c)
        sql_connection.prediction_store.add(c)

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    for c in test_predictions:
        mongo_connection.prediction_store.delete(c.prediction_id)
        postgres_connection.prediction_store.delete(c.prediction_id)
        sql_connection.prediction_store.delete(c.prediction_id)



@pytest.fixture
def options():
    sorting_keys = (None, 
                    "correlation_id", 
                    "timestamp", 
                    "invalid_key"
                    ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"correlation_id":""})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestPredictionStore:

    def test_add_update_delete(self, db_connection: DatabaseConnection, test_prediction: PredictionDTO):
        test_prediction.check_validity()

        read_prediction1 = db_connection.prediction_store.add(test_prediction)
        assert isinstance(read_prediction1.prediction_id, str)
        assert isinstance(read_prediction1.committed_at, datetime.datetime)

        read_prediction1_dict = read_prediction1.to_dict()
        prediction_id = read_prediction1_dict["prediction_id"]
        del read_prediction1_dict["prediction_id"]
        del read_prediction1_dict["committed_at"]

        test_prediction_dict = test_prediction.to_dict()
        del test_prediction_dict["prediction_id"]
        del test_prediction_dict["committed_at"]

        assert read_prediction1_dict == test_prediction_dict

        # Assert we get the same prediction back
        read_prediction2 = db_connection.prediction_store.get(prediction_id)
        assert read_prediction2 is not None
        assert read_prediction2.to_dict() == read_prediction1.to_dict()    

        # Delete the prediction and check that it is deleted
        success = db_connection.prediction_store.delete(prediction_id)
        assert success == True
    

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                gathered_models = db_1.prediction_store.list(*opt, **kwargs)
                count = db_1.prediction_store.count(**kwargs)

                gathered_models2 = db_2.prediction_store.list(*opt, **kwargs)
                count2 = db_2.prediction_store.count(**kwargs)

                assert count == count2
                assert len(gathered_models) == len(gathered_models2)

                for i in range(len(gathered_models)):
                    assert gathered_models[i].prediction_id == gathered_models2[i].prediction_id
                    



import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Dict, Generic, List, Type, TypeVar

import pymongo
from pymongo.database import Database
from sqlalchemy import func, select
from sqlalchemy.orm import Session as SessionClass

from fedn.network.storage.statestore.stores.dto.shared import BaseDTO
from fedn.network.storage.statestore.stores.shared import EntityNotFound
from fedn.network.storage.statestore.stores.sql.shared import MyAbstractBase

MODEL = TypeVar("MODEL", bound=MyAbstractBase)
DTO = TypeVar("DTO", bound=BaseDTO)


def from_document(document: dict) -> dict:
    del document["_id"]
    return document


class Store(ABC, Generic[DTO]):
    """Abstract class for a store.

    OBS! This is an interface, do not add any implementations here.
    """

    @abstractmethod
    def get(self, id: str) -> DTO:
        """Get an entity by id.

        Args:
            id (str): Entity id

        Returns:
            DTO: The entity or null if not found

        """
        pass

    @abstractmethod
    def update(self, item: DTO) -> DTO:
        """Update an existing entity.

        Args:
            item (DTO): The entity to update.

        Returns:
            DTO: The updated entity.

        Raises:
            EntityNotFound: If the entity is not found.
            ValidationError: If validation fails.

        """
        pass

    @abstractmethod
    def add(self, item: DTO) -> DTO:
        """Add an entity.

        Args:
            item (DTO): The entity to update.

        Returns:
            DTO: The updated entity.

        Raises:
            ValidationError: If validation fails.

        """
        pass

    @abstractmethod
    def delete(self, id: str) -> bool:
        """Delete an entity.

        Args:
          id (str): The id of the entity

        Returns:
            Bool: success or failure

        """
        pass

    @abstractmethod
    def list(self, limit: int = 0, skip: int = 0, sort_key: str = None, sort_order=pymongo.DESCENDING, **kwargs) -> List[DTO]:
        """Select entities.

        Args:
            limit (int): The maximum number of entities to return
            skip (int): The number of entities to skip
            sort_key (str): The key to sort by
            sort_order (pymongo.DESCENDING | pymongo.ASCENDING): The order to sort by
            kwargs (dict): Additional query parameters

        Returns:
            List[DTO]: The list of entities

        """
        pass

    @abstractmethod
    def count(self, **kwargs) -> int:
        """Count entities.


        Args:
            kwargs (dict): Additional query parameters, example: {"key": "models"}

        Returns:
            int: The number of entities

        """
        pass


class MongoDBStore(Store[DTO], Generic[DTO]):
    """Base MongoDB store implementation."""

    def __init__(self, database: Database, collection: str, primary_key: str) -> None:
        """Initialize MongoDBStore."""
        self.database = database
        self.collection = collection
        self.primary_key = primary_key
        self.database[self.collection].create_index([(self.primary_key, pymongo.DESCENDING)])

    def get(self, id):
        document = self.database[self.collection].find_one({self.primary_key: id})
        if document is None:
            return None
        return self._dto_from_document(document)

    def add(self, item: DTO) -> DTO:
        item.check_validity(exclude_primary_id=True)
        item_dict = self._document_from_dto(item)

        if self.primary_key not in item_dict or not item_dict[self.primary_key]:
            item_dict[self.primary_key] = str(uuid.uuid4())
        elif self.database[self.collection].find_one({self.primary_key: item_dict[self.primary_key]}):
            raise Exception(f"Entity with id {item_dict[self.primary_key]} already exists")

        item_dict["committed_at"] = datetime.now()

        self.database[self.collection].insert_one(item_dict)
        document = self.database[self.collection].find_one({self.primary_key: item_dict[self.primary_key]})

        return self._dto_from_document(document)

    def update(self, item: DTO) -> DTO:
        raise NotImplementedError("update not implemented for MongoDBStore by default. Use mongo_update in derived classes.")

    def mongo_update(self, item: DTO) -> DTO:
        item.check_validity()
        item_dict = self._document_from_dto(item)
        id = item_dict[self.primary_key]
        result = self.database[self.collection].update_one({self.primary_key: id}, {"$set": item_dict})
        if result.matched_count == 1:
            document = self.database[self.collection].find_one({self.primary_key: id})
            return self._dto_from_document(document)
        raise EntityNotFound(f"Entity with id {id} not found")

    def delete(self, id: str) -> bool:
        result = self.database[self.collection].delete_one({self.primary_key: id})
        return result.deleted_count == 1

    def list(self, limit: int = 0, skip: int = 0, sort_key: str = None, sort_order=pymongo.DESCENDING, **kwargs) -> List[DTO]:
        _sort_order = sort_order or pymongo.DESCENDING
        if sort_key and sort_key != self.primary_key:
            cursor = (
                self.database[self.collection]
                .find(kwargs)
                .sort({sort_key: _sort_order, self.primary_key: pymongo.DESCENDING})
                .skip(skip or 0)
                .limit(limit or 0)
            )
        else:
            cursor = self.database[self.collection].find(kwargs).sort(self.primary_key, pymongo.DESCENDING).skip(skip or 0).limit(limit or 0)

        return [self._dto_from_document(document) for document in cursor]

    def count(self, **kwargs) -> int:
        return self.database[self.collection].count_documents(kwargs)

    @abstractmethod
    def _dto_from_document(self, document: Dict) -> DTO:
        pass

    @abstractmethod
    def _document_from_dto(self, item: DTO) -> Dict:
        pass


class SQLStore(Store[DTO], Generic[DTO, MODEL]):
    """Base SQL store implementation."""

    def __init__(self, Session: Type[SessionClass], SQLModel: Type[MODEL]) -> None:
        """Initialize SQLStore."""
        self.SQLModel = SQLModel
        self.Session = Session

    def get(self, id: str) -> DTO:
        with self.Session() as session:
            stmt = select(self.SQLModel).where(self.SQLModel.id == id)
            entity = session.scalars(stmt).first()
            if entity is None:
                return None
            return self._dto_from_orm_model(entity)

    def add(self, item: DTO) -> DTO:
        item.check_validity(exclude_primary_id=True)
        with self.Session() as session:
            newEntity = self.SQLModel()
            newEntity = self._update_orm_model_from_dto(newEntity, item)

            if not item.primary_id:
                newEntity.id = str(uuid.uuid4())
            else:
                newEntity.id = item.primary_id
            newEntity.committed_at = datetime.now()

            session.add(newEntity)
            session.commit()

            return self._dto_from_orm_model(newEntity)

    def update(self, item: DTO) -> DTO:
        raise NotImplementedError("update not implemented for SQLStore by default. Use sql_update in derived classes.")

    def sql_update(self, item: DTO) -> DTO:
        item.check_validity()
        with self.Session() as session:
            stmt = select(self.SQLModel).where(self.SQLModel.id == item.primary_id)
            existing_item = session.scalars(stmt).first()

            if existing_item is None:
                raise EntityNotFound(f"Entity with id {item.primary_id} not found")

            self._update_orm_model_from_dto(existing_item, item)
            session.commit()

            return self._dto_from_orm_model(existing_item)

    def delete(self, id: str) -> bool:
        with self.Session() as session:
            stmt = select(self.SQLModel).where(self.SQLModel.id == id)
            item = session.scalars(stmt).first()

            if item is None:
                return False

            session.delete(item)
            session.commit()

            return True

    def list(self, limit=0, skip=0, sort_key=None, sort_order=pymongo.DESCENDING, **kwargs) -> List[DTO]:
        with self.Session() as session:
            stmt = select(self.SQLModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(self.SQLModel, key) == value)

            _sort_order = sort_order or pymongo.DESCENDING

            secondary_sort_obj = self.SQLModel.__table__.columns.get("id").desc()
            if sort_key and sort_key in self.SQLModel.__table__.columns:
                sort_obj = self.SQLModel.__table__.columns.get(sort_key)
                if _sort_order == pymongo.DESCENDING:
                    sort_obj = sort_obj.desc()

                stmt = stmt.order_by(sort_obj, secondary_sort_obj)
            else:
                stmt = stmt.order_by(secondary_sort_obj)

            if limit:
                stmt = stmt.offset(skip or 0).limit(limit)
            elif skip:
                stmt = stmt.offset(skip)

            entities = session.scalars(stmt).all()
            return [self._dto_from_orm_model(entity) for entity in entities]

    def count(self, **kwargs) -> int:
        with self.Session() as session:
            stmt = select(func.count()).select_from(self.SQLModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(self.SQLModel, key) == value)

            return session.scalar(stmt)

    @abstractmethod
    def _dto_from_orm_model(self, item: MODEL) -> DTO:
        pass

    @abstractmethod
    def _update_orm_model_from_dto(self, entity: MODEL, item: DTO) -> MODEL:
        pass



import click
import requests

from .main import main
from .shared import CONTROLLER_DEFAULTS, get_api_url, get_response, get_token, print_response


@main.group("session")
@click.pass_context
def session_cmd(ctx):
    """Session commands."""
    pass


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("--n_max", required=False, help="Number of items to list")
@session_cmd.command("list")
@click.pass_context
def list_sessions(ctx, protocol: str, host: str, port: str, token: str = None, n_max: int = None):
    """List sessions."""
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    response = get_response(protocol=protocol, host=host, port=port, endpoint="sessions/", token=token, headers=headers, usr_api=False, usr_token=False)
    print_response(response, "sessions", None)


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("-id", "--id", required=True, help="Session ID")
@session_cmd.command("get")
@click.pass_context
def get_session(ctx, protocol: str, host: str, port: str, token: str = None, id: str = None):
    """Get session by id."""
    response = get_response(protocol=protocol, host=host, port=port, endpoint=f"sessions/{id}", token=token, headers={}, usr_api=False, usr_token=False)
    print_response(response, "session", id)


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("-n", "--name", required=False, help="Name of the session")
@click.option("-a", "--aggregator", required=False, default="fedavg", help="The aggregator plugin to use")
@click.option("-ak", "--aggregator_kwargs", required=False, type=dict, help="Aggregator keyword arguments")
@click.option("-m", "--model_id", required=False, help="The id of the initial model")
@click.option("-rt", "--round_timeout", required=False, default=180, type=int, help="The round timeout to use in seconds")
@click.option("-r", "--rounds", required=False, default=5, type=int, help="The number of rounds to perform")
@click.option("-rb", "--round_buffer_size", required=False, default=-1, type=int, help="The round buffer size to use")
@click.option("-d", "--delete_models", required=False, default=True, type=bool, help="Whether to delete models after each round at combiner (save storage)")
@click.option("-v", "--validate", required=False, default=True, type=bool, help="Whether to validate the model after each round")
@click.option("-hp", "--helper", required=False, help="The helper type to use")
@click.option("-mc", "--min_clients", required=False, default=1, type=int, help="The minimum number of clients required")
@click.option("-rc", "--requested_clients", required=False, default=8, type=int, help="The requested number of clients")
@session_cmd.command("start")
@click.pass_context
def start_session(
    ctx,
    protocol: str,
    host: str,
    port: str,
    token: str,
    name: str = None,
    aggregator: str = "fedavg",
    aggregator_kwargs: dict = None,
    model_id: str = None,
    round_timeout: int = 180,
    rounds: int = 5,
    round_buffer_size: int = -1,
    delete_models: bool = True,
    validate: bool = True,
    helper: str = None,
    min_clients: int = 1,
    requested_clients: int = 8,
):
    """Start a new session."""
    headers = {}
    _token = get_token(token=token, usr_token=False)
    if _token:
        headers = {"Authorization": _token}

    if model_id is None:
        url = get_api_url(protocol, host, port, "models/active", usr_api=False)
        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            model_id = response.json()
        else:
            click.secho(f"Failed to get active model: {response.json()}", fg="red")
            return

    if helper is None:
        url = get_api_url(protocol, host, port, "helpers/active", usr_api=False)
        response = requests.get(url, headers=headers)
        if response.status_code == 400:
            helper = "numpyhelper"
        elif response.status_code == 200:
            helper = response.json()
        else:
            click.secho("An unexpected error occurred when getting the active helper", fg="red")
            return

    url = get_api_url(protocol, host, port, "sessions/", usr_api=False)
    response = requests.post(
        url,
        json={
            "name": name,
            "session_config": {
                "aggregator": aggregator,
                "aggregator_kwargs": aggregator_kwargs,
                "round_timeout": round_timeout,
                "buffer_size": round_buffer_size,
                "model_id": model_id,
                "delete_models_storage": delete_models,
                "clients_required": min_clients,
                "requested_clients": requested_clients,
                "validate": validate,
                "helper_type": helper,
                "server_functions": None,
            },
        },
        headers=headers,
        verify=False,
    )

    if response.status_code == 201:
        session_id = response.json()["session_id"]
        url = get_api_url(protocol, host, port, "sessions/start", usr_api=False)
        response = requests.post(
            url,
            json={
                "session_id": session_id,
                "rounds": rounds,
                "round_timeout": round_timeout,
            },
            headers=headers,
            verify=False,
        )
        response_json = response.json()
        response_json["session_id"] = session_id
        click.secho(f"Session started successfully: {response_json}", fg="green")
    else:
        click.secho(f"Failed to start session: {response.json()}", fg="red")



class EntityNotFound(Exception):
    pass


class ValidationError(Exception):
    def __init__(self, field: str, message: str):
        super().__init__("Validation error on field {}: {}".format(field, message))
        self.field = field
        self.validation_message = message

    def user_message(self):
        return "Validation failed for field '{}' with message: {}".format(self.field, self.validation_message)


class MissingFieldError(Exception):
    def __init__(self, field: str, class_name: str):
        super().__init__("Missing field {} in class {}".format(field, class_name))
        self.field = field
        self.class_name = class_name

    def user_message(self):
        return "Class '{}' is missing field {}".format(self.class_name, self.field)



from typing import Optional

from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, Field, NodeDTO, PrimaryID


class PredictionDTO(BaseDTO):
    """Prediction data transfer object."""

    prediction_id: Optional[str] = PrimaryID(None)
    model_id: str = Field(None)
    data: str = Field(None)
    correlation_id: str = Field(None)
    timestamp: str = Field(None)
    meta: Optional[str] = Field(None)
    sender: NodeDTO = Field(NodeDTO())
    receiver: NodeDTO = Field(NodeDTO())



from typing import Dict

from pymongo.database import Database

from fedn.network.storage.statestore.stores.dto.metric import MetricDTO
from fedn.network.storage.statestore.stores.sql.shared import MetricModel, from_orm_model
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


class MetricStore(Store[MetricDTO]):
    pass


class MongoDBMetricStore(MetricStore, MongoDBStore[MetricDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "metric_id")

    def _document_from_dto(self, item: MetricDTO) -> Dict:
        item_dict = item.to_db(exclude_unset=False)
        return item_dict

    def _dto_from_document(self, document: Dict) -> MetricDTO:
        item = from_document(document)
        return MetricDTO().patch_with(item, throw_on_extra_keys=False)


class SQLMetricStore(MetricStore, SQLStore[MetricDTO, MetricModel]):
    def __init__(self, session):
        super().__init__(session, MetricModel)

    def _update_orm_model_from_dto(self, entity: MetricModel, item: MetricDTO):
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("metric_id", None)
        for key, value in item_dict.items():
            setattr(entity, key, value)
        return entity

    def _dto_from_orm_model(self, item: MetricModel) -> MetricDTO:
        orm_dict = from_orm_model(item, MetricModel)
        orm_dict["metric_id"] = orm_dict.pop("id")
        return MetricDTO().populate_with(orm_dict)
